
import Database.Berkeley.DbMMa

import XmlDatabase
import DatabaseStatistics
import System.Environment

import Text.XML.HaXml.ParseLazy(xmlParse)
import qualified XmlBS 
import qualified ShowXmlLazy as SXL

import qualified Data.Map as Map

import qualified XQuery as XQ
import XQuerySemantics
import XQueryTransform
import MonadExponent
import LazyIO
import TrimSpace

import qualified Data.ByteString.Lazy.Char8 as BSLC8

import XQueryVisualize

renderTypedContent (DocumentC d) = SXL.document (XmlBS.documentFromXmlBS d)
renderTypedContent (ContentC d) = SXL.content (XmlBS.contentFromXmlBS d)
renderTypedContent (BooleanC True) = "true"
renderTypedContent (BooleanC False) = "false"
renderTypedContent (StringC s) = BSLC8.unpack s
renderTypedContent (UntypedAtomicC s) = BSLC8.unpack s

printResult [] = return ()
printResult [x] = putStrLn x
printResult (x:xs) = putStr x >> putStr ", " >> printResult xs

main = do
    (h:pars) <- getArgs
    dbenv <- dbEnv_create []
    dbEnv_open [DB_CREATE,{-DB_INIT_LOCK,DB_INIT_LOG,DB_INIT_TXN,-}DB_INIT_MPOOL{-,DB_RECOVER-}] 0 dbenv (last pars)
    db1 <- db_create [] dbenv
    db_set_re_len db1 13
    db_open [DB_CREATE] DB_QUEUE 0 db1 Nothing "xmlNav.db" Nothing
    db2 <- db_create [] dbenv
    db_set_pagesize db2 32768
    db_open [DB_CREATE] DB_RECNO 0 db2 Nothing "xmlTxt.db" Nothing
    db3 <- db_create [] dbenv
    db_open [DB_CREATE] DB_RECNO 0 db3 Nothing "xmlTag.db" Nothing
    db4 <- db_create [] dbenv
    db_set_pagesize db4 32768
    db_open [DB_CREATE] DB_BTREE 0 db4 Nothing "xmlIdx.db" Nothing
    db5 <- db_create [] dbenv
    db_set_pagesize db5 32768
    db_set_bt_compare_lowlevel db5 int_compare
    db_set_flags db5 [DB_DUP]
--    db_set_flags db5 [DB_DUPSORT]
    db_open [DB_CREATE] DB_BTREE 0 db5 Nothing "xmlParIdx.db" Nothing
    db6 <- db_create [] dbenv
    db_set_flags db6 [DB_DUP]
    db_open [DB_CREATE] DB_BTREE 0 db6 Nothing "xmlTxtIdx.db" Nothing
    let db = Xdbs db1 db2 db3 db4 db5 db6
    case h of
        "load" -> do
            (df,dx) <- case pars of
                [df,_] -> readFile df >>= \x -> return (df, x)
                [_] -> getContents >>= \x -> return ("-", x)
            storeInDB db (XmlBS.documentToXmlBS $ documentTrimSpace $ xmlParse df dx)
--            saveStats (last pars) db
        "query" -> do
--            stats <- loadStats (last pars)
            let [xq, xqx, _] = pars
            d <- readFile xqx
            let res = XQ.doParseModule d
            let (Right (XQ.Module (XQ.Prolog [] [XQ.VarDecl (XQ.VarBinding v _) 
                    (Just (XQ.FunctionCallExpr _ [XQ.StringLiteral df]))]) e)) 
                        = res
--            print e
--            print (map transform e)
--            debug_xquery (XQ.SequenceExpr (map transform e)) stats
            dd <- retrieveFromDB db
            result <- runLazyIO $ runExponentT (evalExpr (map transform e)) (Map.singleton v (DocumentC dd), db)
            printResult (map renderTypedContent result)
        "retrieve" -> do
            dd <- retrieveFromDB db
            putStr (SXL.document (XmlBS.documentFromXmlBS dd))
    db_close [] db1
    db_close [] db2
    db_close [] db3
    db_close [] db4
    db_close [] db5
    db_close [] db6
    dbEnv_close [] dbenv

