module MonadExponent(Exponent(..), ExponentT(..), MonadExponent(..), modifyExponent, unsafeInterleaveExponentIO, unsafeForM) where

import Control.Monad.Trans
import System.IO.Unsafe

class Monad m => MonadExponent s m | m -> s where
    getExponent :: m s
    setExponent :: s -> m a -> m a

newtype Exponent v a = Exponent { runExponent :: v -> a }

instance Monad (Exponent v) where
    return = Exponent . const
    Exponent f >>= g = Exponent (\v -> runExponent (g (f v)) v)

instance MonadExponent v (Exponent v) where
    getExponent = Exponent id
    setExponent v (Exponent f) = Exponent (const (f v))

newtype ExponentT v m a = ExponentT { runExponentT :: v -> m a }

instance Monad m => Monad (ExponentT v m) where
    return = ExponentT . const . return
    ExponentT f >>= g = ExponentT (\v -> f v >>= \w -> runExponentT (g w) v)

instance Monad m => MonadExponent v (ExponentT v m) where
    getExponent = ExponentT return
    setExponent v (ExponentT f) = ExponentT (const (f v))

instance MonadTrans (ExponentT v) where
    lift = ExponentT . const

instance MonadIO m => MonadIO (ExponentT v m) where
    liftIO = ExponentT . const . liftIO

modifyExponent :: MonadExponent s m => (s -> s) -> m a -> m a
modifyExponent f m = getExponent >>= \e -> setExponent (f e) m

unsafeInterleaveExponentIO :: ExponentT v IO a -> ExponentT v IO a
unsafeInterleaveExponentIO f = ExponentT (\v -> unsafeInterleaveIO (runExponentT f v))

unsafeForM :: [a] -> (a -> ExponentT v IO b) -> ExponentT v IO [b]
unsafeForM l f = unsafeInterleaveExponentIO $ fm l where
    fm []    = return []
    fm (h:t) = do
        h' <- unsafeInterleaveExponentIO $ f h
        t' <- unsafeInterleaveExponentIO $ fm t
        return (h':t')

