module RelationAlgebra(evalArith,idxQueryToBExpr,optimize,execute,freevars,num_columns,colsources,usedColumns,usedVars,UsedColumns(..),RelBExpr(..),Subst(..),RelExpr(..),RelTerm(..),Comp(..),Arith(..),RelBAtom(..),FilterAlg(..),JoinAlg(..),IdxQuery(..),Tuple(..),ColType(..),ICO,runICO,renameE,projectE,filterE,cartesianE,andB,orB,eqB,ltB,gtB,leB,geB,trueB,columnT,stringT,addT,subT,intT,varT,sanityCheck,constTableR,rawTableR,joinR,renameR,filterR,projectR) where

import qualified Data.ByteString as BS
import qualified Data.ByteString.Lazy as BSL
import qualified Data.Map as Map
import Data.Map(Map)
import qualified Data.Set as Set
import Data.Set(Set)
import Control.Monad
import Control.Monad.Trans
import Control.Monad.State
import Control.Monad.Reader
import Control.Arrow
import Data.List(elemIndex,partition,sort,nub)
import Data.Maybe(fromJust,maybe)
import XmlDatabase
import Database.Berkeley.DbMMa
import System.IO.Unsafe
import LazyIO
import Binary

--import Debug.Trace
trace _ x = x

type RelBExpr = [[RelBAtom]] -- CNF

data Comp = EqC | NEC | LtC | LEC | GtC | GEC deriving (Show, Eq, Ord)

data Arith = AddA | SubA | MulA | DivA deriving (Show, Eq, Ord)

data RelBAtom = CompB Comp RelTerm RelTerm deriving (Show, Eq, Ord)

data RelTerm = ResultT DbResult
             | VarT String
             | ArithT Arith RelTerm RelTerm
             | ColumnT Int
             deriving (Show, Eq, Ord)

type ICO = Reader [ColName]

compB :: Comp -> ICO RelTerm -> ICO RelTerm -> ICO RelBExpr
compB c = liftM2 (\a b -> [[CompB c a b]])

eqB = compB EqC
neB = compB NEC
ltB = compB LtC
leB = compB LEC
gtB = compB GtC
geB = compB GEC

infix 4 `eqB`, `neB`, `ltB`, `leB`, `gtB`, `geB`

stringT :: BSL.ByteString -> ICO RelTerm
stringT = resultT . DbString

intT :: Int -> ICO RelTerm
intT = resultT . DbInt

resultT :: DbResult -> ICO RelTerm
resultT = return . ResultT

columnT :: ColName -> ICO RelTerm
columnT = liftM ColumnT . columnNum

columnNumT :: Int -> ICO RelTerm
columnNumT = return . ColumnT

varT :: String -> ICO RelTerm
varT = return . VarT

arithT :: Arith -> ICO RelTerm -> ICO RelTerm -> ICO RelTerm
arithT a = liftM2 (ArithT a)

addT = arithT AddA
subT = arithT SubA

infixl 6 `addT`
infixl 6 `subT`

andB :: ICO RelBExpr -> ICO RelBExpr -> ICO RelBExpr
andB = liftM2 (++)

infixr 3 `andB`

orB :: ICO RelBExpr -> ICO RelBExpr -> ICO RelBExpr
orB = error "orB unimpl"

infixr 2 `orB`

trueB :: ICO RelBExpr
trueB = return []

falseB :: ICO RelBExpr
falseB = return [[]]

fromJustX x (Just a) = a
fromJustX x Nothing = error x

columnNum :: ColName -> ICO Int
columnNum n = liftM (fromJustX "columnNum" . elemIndex n) ask

data JoinAlg = NestedLoopJoin RelBExpr | MergeJoin [(Int,Int)] | IndexLoopJoin [Int] (IdxQuery RelTerm)
    deriving Show

data FilterAlg = SimpleFilter RelBExpr | IndexFilter [Int] (IdxQuery RelTerm)
    deriving Show

data IdxQuery a = QSingle [a] | QRange (Maybe [a]) (Maybe [a])
    deriving Show

instance Functor IdxQuery where
    fmap f (QSingle l) = QSingle (fmap f l)
    fmap f (QRange m1 m2) = QRange (fmap (fmap f) m1) (fmap (fmap f) m2)

data RelInfo = RelInfo { 
    ifreevars    :: [String], 
    icolsources  :: [Maybe (String,Int)],
    icoltypes    :: [ColType],
    inum_columns :: Int, 
    iindexes     :: [[Int]] }
    deriving Show

data RelExpr = RawTable   RelInfo String String 
             | ConstTable RelInfo [ColName] [ColType] [TermTuple] String
             | Rename     RelInfo [ColName] String RelExpr
             | Join       RelInfo JoinAlg RelExpr RelExpr
             | Filter     RelInfo FilterAlg RelExpr
             | Project    RelInfo [Int] RelExpr
             | Limit      RelInfo Int RelExpr
--             | Calculate [ColName] [ColType] [RelTerm] String RelExpr
--             | Union RelExpr RelExpr
--             | Distinct RelExpr
             deriving Show

data ColType = IntType | VarcharType deriving Show

type ColName = String

encodeDbResult :: DbResult -> BS.ByteString
encodeDbResult (DbBool True) = encodeDbResult (DbInt 1)
encodeDbResult (DbBool False) = encodeDbResult (DbInt 0)
encodeDbResult (DbInt i) = encodeKey i
encodeDbResult (DbString s) = BS.concat $ BSL.toChunks s

decodeDbResult :: ColType -> BS.ByteString -> DbResult
decodeDbResult IntType s = DbInt $ decodeKey s
decodeDbResult VarcharType s = DbString $ BSL.fromChunks [s]

type Tuple = [DbResult]
type TermTuple = [RelTerm]

type Subst = Map.Map String DbResult

x `usesColumnsFrom` e = 
    usedColumns x `Set.isSubsetOf` Set.fromAscList [0..num_columns e-1]

sanityCheck (RawTable _ _ _) = True
sanityCheck (ConstTable _ n t ts nm) = 
    length n == length t && all ((== length n) . length) ts
sanityCheck (Rename _ ns n e) = length ns == num_columns e && sanityCheck e
sanityCheck e@(Join _ (NestedLoopJoin b) e1 e2) = 
    b `usesColumnsFrom` e && sanityCheck e1 && sanityCheck e2
sanityCheck (Join _ (IndexLoopJoin i q) e1 e2) =
    renumber (+(-num_columns e1)) i `elem` indexes e2 && q `usesColumnsFrom` e1 &&
    sanityCheck e1 && sanityCheck e2
sanityCheck (Filter _ (SimpleFilter b) e) =
    b `usesColumnsFrom` e && sanityCheck e
sanityCheck (Filter _ (IndexFilter i q) e) =
    i `elem` indexes e && Set.null (usedColumns q) && sanityCheck e
sanityCheck (Project _ c e) = c `usesColumnsFrom` e && sanityCheck e
sanityCheck (Limit _ k e) = sanityCheck e

sanityCheckS s e | sanityCheck e = e
                 | otherwise     = error ("Sanity check failed:" ++ s)

fullname :: String -> String -> String
fullname tn fn = tn ++ "." ++ fn

info (RawTable i _ _) = i
info (ConstTable i _ _ _ _) = i
info (Rename i _ _ _) = i
info (Join i _ _ _) = i
info (Filter i _ _) = i
info (Project i _ _) = i
info (Limit i _ _) = i

freevars = ifreevars . info
num_columns = inum_columns . info
indexes = iindexes . info
coltypes = icoltypes . info
colsources = icolsources . info

rawTableR x n = RawTable i x n where
    i = case x of
        "nav" -> RelInfo {
            ifreevars = map (uncurry fullname) [(n,"pre"),(n,"parent"),(n,"size"),(n,"type"),(n,"tag")],
            icolsources = map Just $ zip (repeat "nav") [0..4],
            icoltypes = [IntType, IntType, IntType, IntType, IntType],
            inum_columns = 5,
            iindexes = [[0],[1]]
        }
        "txt" -> RelInfo {
            ifreevars = map (uncurry fullname) [(n,"pre"),(n,"text")],
            icolsources = map Just $ zip (repeat "txt") [0,1],
            icoltypes = [IntType, VarcharType],
            inum_columns = 2,
            iindexes = [[0],[1]]
        }
        "tag" -> RelInfo {
            ifreevars = map (uncurry fullname) [(n,"tag"),(n,"ttext")],
            icolsources = map Just $ zip (repeat "tag") [0,1],
            icoltypes = [IntType, VarcharType],
            inum_columns = 2,
            iindexes = [[0],[1]]
        }
        "idx_parent" -> RelInfo {
            ifreevars = map (uncurry fullname) [(n,"parent"),(n,"pre")],
            icolsources = map Just $ zip (repeat "txt") [0,1],
            icoltypes = [IntType, IntType],
            inum_columns = 2,
            iindexes = [[0]]
        }
        "idx_text" -> RelInfo {
            ifreevars = map (uncurry fullname) [(n,"text"),(n,"pre")],
            icolsources = map Just $ zip (repeat "txt") [0,1],
            icoltypes = [VarcharType, IntType],
            inum_columns = 2,
            iindexes = [[0]]
        }
        "idx_ttext" -> RelInfo {
            ifreevars = map (uncurry fullname) [(n,"ttext"),(n,"tag")],
            icolsources = map Just $ zip (repeat "tag") [0,1],
            icoltypes = [VarcharType, IntType],
            inum_columns = 2,
            iindexes = [[0]]
        }

constTableR fs tys ts n = ConstTable i fs tys ts n where
    i = RelInfo {
        ifreevars = map (fullname n) fs,
        icolsources = map (const Nothing) fs,
        icoltypes = tys,
        inum_columns = length fs,
        iindexes = []
    }

renameR fs n e = Rename i fs n e where
    i = RelInfo {
        ifreevars = map (fullname n) fs,
        icolsources = colsources e,
        icoltypes = coltypes e,
        inum_columns = num_columns e,
        iindexes = indexes e
    }

joinR b e1 e2 = Join i b e1 e2 where
    i = RelInfo {
        ifreevars = freevars e1 ++ freevars e2,
        icolsources = colsources e1 ++ colsources e2,
        icoltypes = coltypes e1 ++ coltypes e2,
        inum_columns = num_columns e1 + num_columns e2,
        iindexes = indexes e1
    }

filterR b e = Filter i b e where
    i = RelInfo {
        ifreevars = freevars e,
        icolsources = colsources e,
        icoltypes = coltypes e,
        inum_columns = num_columns e,
        iindexes = indexes e
    }

projectR s e = Project i s e where
    i = RelInfo {
        ifreevars = rearrange s $ freevars e,
        icolsources = rearrange s $ colsources e,
        icoltypes = rearrange s $ coltypes e,
        inum_columns = length s,
        iindexes = map (map $ reverseProject s) . filter (all (`elem` s)) $ indexes e
    }

limitR k e = Limit i k e where
    i = RelInfo {
        ifreevars = freevars e,
        icolsources = colsources e,
        icoltypes = coltypes e,
        inum_columns = num_columns e,
        iindexes = indexes e
    }

--freevars (Calculate fs _ _ n _) = map (fullname n) fs
--freevars (Union e1 e2) = freevars e1

data WOrder a = AscOrd a | DescOrd a

instance Functor WOrder where
    fmap f (AscOrd a) = AscOrd (f a)
    fmap f (DescOrd a) = DescOrd (f a)

ordering :: RelExpr -> [WOrder Int]
ordering (RawTable _ _ _) = [DescOrd 0]
ordering (ConstTable _ _ _ _ _) = []
--ordering (Calculate _ _ _ _ _) = []
ordering (Rename _ _ _ e) = ordering e
ordering (Join _ _ e1 e2) = ordering e1 -- ++ ordering e2  (przenumerowane!)
ordering (Filter _ _ e) = ordering e
ordering (Project _ s e) = [] -- pomyśleć
--ordering (Union _ _) = []
ordering (Limit _ _ e) = ordering e

-- smart constructory

runICO :: RelExpr -> ICO a -> a
runICO re i = runReader i (freevars re)

runICO_ i = runReader i []

filterE :: ICO RelBExpr -> RelExpr -> RelExpr
filterE b e = filterR (SimpleFilter $ runICO e b) e

projectE :: [ColName] -> RelExpr -> RelExpr
projectE fs e = projectR (runICO e $ mapM columnNum fs) e 

renameE = renameR

joinNLJ :: ICO RelBExpr -> RelExpr -> RelExpr -> RelExpr
joinNLJ b e1 e2 = 
    joinR (NestedLoopJoin $ runICO (cartesianE e1 e2) b) e1 e2

cartesianE = joinR (NestedLoopJoin [])

--

evalComp EqC = (==)
evalComp NEC = (/=)
evalComp LtC = (<)
evalComp LEC = (<=)
evalComp GtC = (>)
evalComp GEC = (>=)

evalArith AddA = (+)
evalArith SubA = (-)
evalArith MulA = (*)

evalTerm :: Tuple -> Subst -> RelTerm -> DbResult
evalTerm _ _ (ResultT r) = r
evalTerm t _ (ColumnT k) = t !! k
evalTerm _ s (VarT n) = fromJustX "evalTerm" $ Map.lookup n s
evalTerm t s (ArithT a t1 t2) = evalArith a (evalTerm t s t1) (evalTerm t s t2)

evalBAtom :: Tuple -> Subst -> RelBAtom -> Bool
evalBAtom t s (CompB c t1 t2) = evalComp c (evalTerm t s t1) (evalTerm t s t2)

evalBExpr :: Tuple -> Subst -> RelBExpr -> Bool
evalBExpr t s = and . map (or . map (evalBAtom t s)) 

-- przenumerowania, przydatne do przekształceń
-- używane kolumny, przydatne do analiz

class UsedColumns a where
    usedColumns :: a -> Set Int
    usedVars    :: a -> Set String
    renumber    :: (Int -> Int) -> a -> a
    subst       :: (String -> RelTerm) -> a -> a

instance UsedColumns Int where
    usedColumns = Set.singleton
    usedVars k  = Set.empty
    renumber f  = f
    subst _     = id

instance UsedColumns a => UsedColumns [a] where
    usedColumns = Set.unions . map usedColumns
    usedVars    = Set.unions . map usedVars
    renumber f  = fmap (renumber f)
    subst f     = fmap (subst f)

instance UsedColumns a => UsedColumns (Maybe a) where
    usedColumns = maybe Set.empty usedColumns
    usedVars    = maybe Set.empty usedVars
    renumber f  = fmap (renumber f)
    subst f     = fmap (subst f)

instance UsedColumns FilterAlg where
    usedVars (SimpleFilter b) = usedVars b
    usedVars (IndexFilter _ iq) = usedVars iq
    usedColumns (SimpleFilter b) = usedColumns b
    usedColumns (IndexFilter cs iq) = Set.fromList cs
    renumber f (SimpleFilter b) = SimpleFilter (renumber f b)
    renumber f (IndexFilter cs iq) = IndexFilter (map f cs) (renumber f iq)
    subst f (SimpleFilter b) = SimpleFilter (subst f b)
    subst f (IndexFilter cs iq) = IndexFilter cs (subst f iq)

instance UsedColumns JoinAlg where
    usedVars (NestedLoopJoin b)   = usedVars b
    --usedVars (MergeJoin _)        = Set.empty
    usedVars (IndexLoopJoin _ iq) = usedVars iq
    usedColumns (NestedLoopJoin b) = usedColumns b
    usedColumns (IndexLoopJoin i q) = usedColumns i `Set.union` usedColumns q
    renumber f (NestedLoopJoin b) = NestedLoopJoin (renumber f b)
    renumber f (IndexLoopJoin i q) = IndexLoopJoin (renumber f i) (renumber f q)

instance UsedColumns a => UsedColumns (IdxQuery a) where
    usedVars (QSingle l)    = usedVars l
    usedVars (QRange l1 l2) = usedVars l1 `Set.union` usedVars l2
    usedColumns (QSingle l)    = usedColumns l
    usedColumns (QRange l1 l2) = usedColumns l1 `Set.union` usedColumns l2
    renumber f (QSingle l)    = QSingle (renumber f l)
    renumber f (QRange l1 l2) = QRange (renumber f l1) (renumber f l2)

instance UsedColumns RelExpr where
    usedVars (ConstTable _ _ _ ts _) = usedVars ts
    usedVars (RawTable _ _ _) = Set.empty
    usedVars (Filter _ b e)   = usedVars b  `Set.union` usedVars e
--    usedVars (Union e1 e2)  = usedVars e1 `Set.union` usedVars e2
    usedVars (Join _ b e1 e2) = Set.unions [usedVars b, usedVars e1, usedVars e2]
    usedVars (Rename _ _ _ e) = usedVars e
    usedVars (Limit _ _ e)    = usedVars e
    usedVars (Project _ _ e)  = usedVars e

instance UsedColumns RelBAtom where
    usedColumns (CompB _ t1 t2) = usedColumns t1 `Set.union` usedColumns t2
    usedVars    (CompB _ t1 t2) = usedVars    t1 `Set.union` usedVars    t2
    renumber f  (CompB c t1 t2) = CompB c (renumber f t1) (renumber f t2)
    subst f     (CompB c t1 t2) = CompB c (subst f t1) (subst f t2)

instance UsedColumns RelTerm where
    usedColumns (ColumnT k) = Set.singleton k
    usedColumns (ArithT _ t1 t2) = usedColumns t1 `Set.union` usedColumns t2
    usedColumns _ = Set.empty
    usedVars    (VarT k) = Set.singleton k
    usedVars    (ArithT _ t1 t2) = usedVars t1 `Set.union` usedVars t2
    usedVars    _ = Set.empty
    renumber f (ColumnT k) = ColumnT (f k)
    renumber f (ArithT a t1 t2) = ArithT a (renumber f t1) (renumber f t2)
    renumber f t = t
    subst f (VarT v) = f v
    subst f (ArithT a t1 t2) = ArithT a (subst f t1) (subst f t2)
    subst f t = t

--

getSubst' = liftM snd ask
getDb' = liftM fst ask
        
{-# INLINE dbread #-}
dbread "nav" k v = 
    [k, DbInt prn, DbInt size, DbInt (fromIntegral typ), DbInt tag]
    where nd@(NavData typ size tag prn) = decodeNav v
dbread "txt" k v = 
    [k, DbString (BSL.fromChunks [v])]
dbread "tag" k v = 
    [k, DbString (BSL.fromChunks [v])]
dbread ('i':'d':'x':_) k v = 
    [k, DbInt (decodeKey v)]

{-# INLINE sizes #-}
sizes "nav" = (navDb, Just 4, Just 13)
sizes "txt" = (txtDb, Just 4, Nothing)
sizes "tag" = (tagDb, Just 4, Nothing)
sizes "idx_parent" = (parentIdx, Just 4, Just 4)
sizes "idx_text" = (txtIdx, Nothing, Just 4)
sizes "idx_ttext" = (txtIdx, Nothing, Just 4)

--loopjoin :: (Monad m) => ([a] -> [a] -> m [a]) -> m [[a]] -> ([a] -> m [[a]]) -> ([a] -> [a] -> m [a]) -> m [a]
--{-# INLINE loopjoin #-}
--loopjoin fin src1 src2 hf = liftM concat . (src1 >>=) . mapM $ \t1 -> do
--    (hf t1 =<<) . liftM concat . (src2 t1 >>=) . mapM $ \t2 -> fin t1 t2
--loopjoin fin src1 src2 hf = liftM concat . (src1 >>=) . mapM $ \t1 -> do
--    (hf t1 =<<) . liftM concat . (src2 t1 >>=) . mapM $ \t2 -> fin t1 t2
{-
loopjoin fin src1 src2 = liftM concat . (src1 >>=) . mapM $ \t1 -> do
    liftM concat . (src2 t1 >>=) . mapM $ fin t1
-}
loopjoin fin src1 src2 = do
    x <- src1
    plj x 
    where
    plj [] = return []
    plj x@(hx:tx) = do
        y <- src2 hx
        lj' x y
    lj' [_] [] = return []
    lj' (_:x@(hx:tx)) [] = do
        y <- src2 hx
        lj' x y
    lj' x@(hx:_) (hy:ty) = do
        v <- fin hx hy
        t <- lj' x ty
        return (v ++ t)
    lj' x y = error ("fuck: " ++ show (length x) ++ " " ++ show(length y))

{-
loopjoin fin src1 src2 = do
    x <- src1
    flip concatMapM x $ \t1 -> do
        y <- src2 t1
        concatMapM (fin t1) y
-}      
--concatMapM f = liftM concat . mapM f
{-
concatMapM f [] = return []
concatMapM f (h:t) = do
    d <- f h
    r <- concatMapM f t
    return (d ++ r)
-}

idxQueryToBExpr k (QSingle v) = runICO_ $ foldr1 andB $ 
    zipWith eqB (map columnNumT k) (map return v)
idxQueryToBExpr [k] (QRange r1 r2) = runICO_ $ h leB r1 `andB` h geB r2 where
    h i Nothing = trueB
    h i (Just [r]) = return r `i` columnNumT k

complex cf (Filter _ (SimpleFilter b) e) = {-# SCC "complex_Filter" #-} do
    s <- getSubst'
    liftM (filter $ \t -> evalBExpr t s b) $ cf e
complex cf (Project _ s e) = {-# SCC "complex_Project" #-} liftM (map $ rearrange s) $ cf e
--complex cf (Union e1 e2) = liftM2 (++) (cf e1) (cf e2)
complex cf (Rename _ _ _ e) = {-# SCC "complex_Rename" #-} cf e
complex cf (Limit _ k e) = {-# SCC "complex_Limit" #-} liftM (take k) $ cf e
--complex cf (Calculate _ _ tc _ e) = do
--    s <- getSubst'
--    mapM (\t -> return $ map (evalTerm t s) tc) =<< cf e
complex cf (Join _ (NestedLoopJoin cond) e1 e2) = {-# SCC "complex_NLJ" #-} 
    loopjoin fin (cf e1) (const $ execute e2)
    where
    fin t1 t2 = do
        let t = t1 ++ t2
        s <- getSubst'
        return $ (guard $ evalBExpr t s cond) >> return t
complex cf (Join _ (IndexLoopJoin idx iq) e1 e2) = {-# SCC "complex_ILJ" #-}
    loopjoin fin (cf e1) (xf e2)
    where
    fin t1 t2 = return . return $ t1 ++ t2
    xf e2 = trf >=> \t -> idxlookup (renumber (+(-num_columns e1)) idx) t e2
    trf t = do
        s <- getSubst'
        return $ fmap (evalTerm t s) iq
complex cf (Join _ (MergeJoin jps) e1 e2) = {-# SCC "complex_MJ" #-}
    liftM2 merge (cf e1) (execute e2) 
    where
    merge :: [Tuple] -> [Tuple] -> [Tuple]
    merge [] _ = []
    merge _ [] = []
    merge l1@(t1:ts1) l2@(t2:ts2) 
        | c == EQ = [u1 ++ u2 | u1 <- t1:ul1, u2 <- t2:ul2] ++ merge us1 us2
        | c == LT = merge ts1 l2
        | c == GT = merge l1 ts2
        where 
        k1 = rearrange (map fst jps) t1
        k2 = rearrange (map snd jps) t2
        c = compare k1 k2
        (ul1, us1) = break (\u1 -> rearrange (map fst jps) u1 /= k1) ts1
        (ul2, us2) = break (\u2 -> rearrange (map snd jps) u2 /= k2) ts2
complex cf x = error ("complex: " ++ show x)

execute :: RelExpr -> ReaderT (Xdbs, Subst) LazyIO [Tuple]
execute r@(RawTable _ n _) = idxlookup [0] (QRange Nothing Nothing) r
execute (ConstTable _ _ _ ts _) = do
    s <- getSubst'
    return $ map (map $ evalTerm [] s) ts
execute (Filter _ (IndexFilter cs iq) e) = do
    s <- getSubst'
    idxlookup cs (fmap (evalTerm [] s) iq) e
execute e = do
    r <- complex execute e
    return $ trace ("execute " ++ show e ++ "\n" ++ show r) r

idxlookup :: [Int] -> IdxQuery DbResult -> RelExpr -> ReaderT (Xdbs, Subst) LazyIO [Tuple]
idxlookup i cf (Project _ s e) = liftM (map $ rearrange s) $ idxlookup (map (s!!) i) cf e
idxlookup i (QSingle v) t@(RawTable _ _ _) = 
    idxlookup i (QRange (Just v) (Just v)) t
idxlookup [0] q@(QRange r1 r2) e@(RawTable _ n _) = do
    db <- getDb'
    r <- liftIO $ initScan r1 =<< db_cursor [] (dbsel db) Nothing
    return $ trace ("idxlookup " ++ show [0] ++ " " ++ show q ++ " " ++ show e ++ "\n" ++ show r) r
    where 
    decodeIdxKey = decodeDbResult (coltypes e !! 0)
    initScan Nothing    = seqScan
    initScan (Just [k]) = funScan (\dbc ->
        dbc_get_range_sz ksize vsize [] dbc (encodeDbResult k))
    seqScan = funScan (\dbc -> dbc_move_get_sz ksize vsize [] dbc DB_NEXT)
    funScan fun dbc = unsafeInterleaveIO $ do
        r <- fun dbc
        case r of
            Nothing -> dbc_close dbc >> return []
            Just (k, v) -> kguard r2 (decodeIdxKey k) $ do
                liftM (dbread n (decodeIdxKey k) v:) (seqScan dbc)
    kguard Nothing           _ m = m
    kguard (Just [k']) k m | k > k'    = return []
                           | otherwise = m
    (dbsel, ksize, vsize) = sizes n
idxlookup [1] iq (RawTable _ "nav" _) = loopjoin fin cf xf
    where
    fin _ = return . return 
    cf = idxlookup [0] iq (rawTableR "idx_parent" "idx")
    xf = \[_, pre] -> idxlookup [0] (QSingle [pre]) (rawTableR "nav" "nav")
idxlookup [1] iq (RawTable _ "txt" _) = liftM (map $ \[a,b] -> [b,a]) $ 
    idxlookup [0] iq (rawTableR "idx_text" "idx")
idxlookup [1] iq (RawTable _ "tag" _) = liftM (map $ \[a,b] -> [b,a]) $ 
    idxlookup [0] iq (rawTableR "idx_ttext" "idx")
idxlookup i q (Filter _ (IndexFilter cs iq) e) = do
    s <- getSubst'
    let b = idxQueryToBExpr cs (fmap (ResultT . evalTerm [] s) iq)
    idxlookup i q (filterR (SimpleFilter b) e)
idxlookup n q e =  do
    r <- complex (idxlookup n q) e
    return $ trace ("idxlookup " ++ show n ++ " " ++ show q ++ " " ++ show e ++ "\n" ++ show r) r

rearrange s t = map (t!!) s

--peepholeIterator h (Calculate ns ty t n e) = liftM (Calculate ns ty t n) (h e)
peepholeIterator h (Rename _ x y e)  = liftM (renameR x y) (h e)
peepholeIterator h (Project _ s e)   = liftM (projectR s) (h e)
--peepholeIterator h (Union e1 e2)   = liftM2 Union (h e1) (h e2)
peepholeIterator h (Limit _ k e)     = liftM (limitR k) (h e)
peepholeIterator h (Filter _ b e)    = liftM (filterR b) (h e)
peepholeIterator h (Join _ ja e1 e2) = liftM2 (joinR ja) (h e1) (h e2)
peepholeIterator h e               = return e

iteratePeepholeUp :: (RelExpr -> Maybe RelExpr) -> RelExpr -> Maybe RelExpr
iteratePeepholeUp f e = if c then Just e' else Nothing where
    (e', c) = runState (h e) False
    h e = case f e of
        Just e' | sanityCheck e' -> put True >> h e'
                | otherwise      -> error "iteratePeepholeUp"
        Nothing -> peepholeIterator h e

iteratePeepholeDown :: (RelExpr -> Maybe RelExpr) -> RelExpr -> Maybe RelExpr
iteratePeepholeDown f e = if c then Just e' else Nothing where
    (e', c) = runState (h e) False 
    h e = peepholeIterator h e >>= g 
    g e = case f e of
        Just e' | sanityCheck e' -> put True >> g e'
                | otherwise      -> error "iteratePeepholeDown"
        Nothing -> return e

splitBExpr :: [Set Int] -> RelBExpr -> (RelBExpr, [RelBExpr])
splitBExpr evs = foldr h ([], map (const []) evs) where
    h c = h' (usedColumns c) c evs 
    h' cv c []       (b', []  )  = (c:b', [])
    h' cv c (ev:evs) (b', b:bs) 
        | cv `Set.isSubsetOf` ev = (b', (c:b):bs)
        | otherwise              = second (b:) (h' cv c evs (b', bs))

shiftFiltersUp = iteratePeepholeUp sf where
    sf (Filter _ (SimpleFilter []) e) = return e
    sf (Filter _ (SimpleFilter b) (Join _ ja e1 e2))
        | null b1 && null b2 = mzero
        | otherwise = return $ filterR (SimpleFilter b') $ joinR ja 
            (filterR (SimpleFilter b1) e1) (filterR (SimpleFilter $ renumber (+(-e1nc)) b2) e2)
            where
            e1nc = num_columns e1
            e2nc = num_columns e2
            e1v = Set.fromDistinctAscList [0 .. e1nc - 1]
            e2v = Set.fromDistinctAscList [e1nc .. e1nc + e2nc - 1]
            (b', [b1, b2]) = splitBExpr [e1v, e2v] b
    sf (Join _ (NestedLoopJoin b@(_:_)) e1 e2) =
        return $ filterR (SimpleFilter b) (joinR (NestedLoopJoin []) e1 e2)
    sf (Filter _ (SimpleFilter b1) (Filter _ (SimpleFilter b2) e)) = 
        return $ filterR (SimpleFilter $ b1 ++ b2) e
    sf (Filter _ b (Rename _ x y e)) = 
        return $ renameR x y (filterR b e)
    sf (Filter _ b (Project _ s e)) = 
        return $ projectR s (filterR (renumber (s!!) b) e)
--    sf (Filter b (Union e1 e2)) = return $ Union (Filter b e1) (Filter b e2)
--    sf (Filter b (Limit k e)) = return $ Limit k (Filter b e)
    sf _ = mzero

reverseProject l x = fromJustX ("reverseProject " ++ show l ++ show x) . flip Map.lookup (Map.fromList (zip l [0..])) $ x

shiftProjectionsUp = iteratePeepholeUp sp where
    sp (Project _ l e) | l == [0..num_columns e-1] = return e
    sp (Project _ l (Join _ b e1 e2)) 
        | sort rql == [0..e1nc+e2nc-1] = mzero
        | otherwise = return $ projectR (map r l) $ joinR (renumber r b) (projectR l1 e1) (projectR (map (+(-e1nc)) l2) e2)
            where
            r = reverseProject (l1 ++ l2)
            rql = nub $ l ++ Set.toList (usedColumns b)
            e1nc = num_columns e1
            e2nc = num_columns e2
            (l1, l2) = partition (< e1nc) rql
    sp (Project _ l (Project _ m e)) = return $ projectR (map (m!!) l) e
    sp (Project _ l (Filter _ b e)) 
        | sort rql == [0..enc-1] = mzero
        | otherwise = return $ projectR (map r l) $ filterR (renumber r b) $ projectR rql e
            where 
            enc = num_columns e
            rql = nub $ l ++ Set.toList (usedColumns b)
            r = reverseProject rql
    sp (Project _ l (Rename _ x n e)) = return $ renameR (map (x!!) l) n (projectR l e)
--    sp (Project l (Union e1 e2)) = return $ Union (Project l e1) (Project l e2)
    sp (Project _ l (Limit _ k e)) = return $ limitR k (projectR l e)
    sp _ = mzero

integrateFilterJoin = iteratePeepholeUp sf where
    sf (Filter _ (SimpleFilter b) (Join _ (NestedLoopJoin b') e1 e2)) =
        return $ joinR (NestedLoopJoin (b ++ b')) e1 e2
    sf _ = mzero

flippedComp EqC = EqC
flippedComp NEC = NEC
flippedComp LtC = GtC
flippedComp LEC = GEC
flippedComp GtC = LtC
flippedComp GEC = LEC

reorientBAtom (CompB comp t1 t2) | t1 < t2   = CompB comp t1 t2
                                 | otherwise = CompB (flippedComp comp) t2 t1

reorientBExpr = sort . map (sort . map reorientBAtom)

promoteNLJtoILJ = iteratePeepholeDown sf where
    sf (Filter _ (SimpleFilter []) e) = return e
    sf (Filter _ (SimpleFilter b) e) = do
        Filter _ (SimpleFilter b') (Join _ (IndexLoopJoin k q) (ConstTable _ [] [] [[]] "lol") e') <- sf (joinR (NestedLoopJoin b) (constTableR [] [] [[]] "lol") e)
        return $ filterR (SimpleFilter b') $ filterR (IndexFilter k q) e'
    sf j@(Join _ (NestedLoopJoin b) e1 e2) 
        | (r, k, q):_ <- psbls = {-trace ("ILJ " ++ show splits) $ -}
            return $ filterR (SimpleFilter r) $ joinR (IndexLoopJoin k q) e1 e2
        | otherwise = {-trace ("ILJ Failure: " ++ show j) $-} mzero 
        where
        e1nc = num_columns e1
        rb = reorientBExpr b
        idxs = map (id &&& Set.fromList . (++[0..e1nc-1])) . 
               map (map (+e1nc)) $ indexes e2
        splits = map (id *** id *** head) . filter (not . null . snd . snd) . 
                   map (id *** flip splitBExpr rb . return) $ idxs
        psbls = msum (map h splits)
        h a@([k],_) | trace ("SHITSHITLOLOLOL" ++ show k ++ " " ++ show a) False = undefined
        h ([k], (r, [[CompB EqC t1 (ColumnT k')]])) 
            | k == k' && t1 `usesColumnsFrom` e1 = 
                trace "A" $ [(r, [k], QSingle [t1])]
        h ([k], (r, [[CompB LEC t1 (ColumnT k1)],[CompB GEC t2 (ColumnT k2)]]))
            | k == k1 && k == k2 && 
              t1 `usesColumnsFrom` e1 && t2 `usesColumnsFrom` e1 =
                trace "B" $ [(r, [k], QRange (Just [t1]) (Just [t2]))]
        h a = []
    sf _ = mzero

cleanupRel = iteratePeepholeDown cr where
    cr (Filter _ (SimpleFilter []) e) = return e
    cr (Project _ l (Project _ m e)) = return $ projectR (map (m!!) l) e
    cr (Project _ l e) | l == [0..num_columns e-1] = return e
    cr (Join _ (NestedLoopJoin []) (ConstTable _ [] [] [_] _) e) = return e
    cr (Join _ (NestedLoopJoin []) e (ConstTable _ [] [] [_] _)) = return e
--    cr (Project _ l (Join _ (IndexLoopJoin [1] (QSingle [_])) (Project _ [0] (RawTable _ _ _)) e)) 
--        | 0 `notElem` l = return (projectR (renumber (+(-1)) l) e)
    cr (Project _ l (Join _ (IndexLoopJoin [_] (QSingle [_])) e (Project _ [0] (RawTable _ _ _)))) 
        | num_columns e `notElem` l = return (projectR l e)
    cr (Filter _ (IndexFilter [0] (QSingle [e@(VarT _)])) (Project _ [0] (RawTable _ _ _))) =
        return $ constTableR ["key"] [IntType] [[e]] "lol"
    cr _ = mzero

hackLimitUnused = iteratePeepholeUp lu where
    lu (Project _ [] e) | Limit _ 1 _ <- e = mzero
                        | otherwise = return $ projectR [] (limitR 1 e)
    lu (Project _ l@(_:_) (Join _ a@(IndexLoopJoin _ _) e1 e2)) 
        | Limit _ 1 _ <- e2 = mzero
        | all (< num_columns e1) l = return $ projectR l (joinR a e1 (limitR 1 e2))
    lu _ = mzero

shiftAdminJoinsUp = iteratePeepholeDown saj where
--    saj (Project _ l (Join _ a e1 (Limit _ 1 e2)))
    saj (Project _ l (Join _ a z@(Project _ m e1) (Limit _ 1 e2))) 
        | all (< num_columns z) l =
        return $ projectR (map f l) (joinR (renumber f a) e1 (limitR 1 e2))
        where
        f x | x < num_columns z = m !! x
            | otherwise = x - num_columns z + num_columns e1
    saj (Project _ l (Join _ a z@(Join _ b e1a e1b) (Limit _ 1 e2))) 
        | Limit _ _ _ <- e1b = mzero
        | (Set.null $ usedColumns a `Set.intersection` Set.fromList [num_columns e1a..num_columns z-1]) && all (< num_columns z) l = 
        return $ projectR l (joinR b (projectR [0..num_columns e1a-1] (joinR (renumber f a) e1a (limitR 1 e2))) e1b)
        where
        f x | x < num_columns e1a = x
            | otherwise = x - num_columns z + num_columns e1a
    saj _ = mzero

demaybize f e = maybe e id (f e)

iteratetrans 0 _ e = e
iteratetrans k f e = maybe e (iteratetrans (k-1) f) (f e)

optimize = iteratetrans 100 shiftAdminJoinsUp . demaybize hackLimitUnused . demaybize shiftProjectionsUp . demaybize cleanupRel . demaybize shiftProjectionsUp . demaybize cleanupRel . demaybize promoteNLJtoILJ . demaybize shiftProjectionsUp . demaybize integrateFilterJoin . demaybize shiftFiltersUp

