module XQueryTransform where

import Control.Monad
import Control.Monad.State

import Control.Arrow
import qualified XQuery as XQ
import XQuerySemantics
import RelationAlgebra
import qualified Data.ByteString.Lazy.Char8 as BSLC8
import qualified Data.Set as Set
import Data.List(elemIndex,find)
import Debug.Trace(trace)
import XmlDatabase(DbResult(..))

fullxml :: RelExpr
fullxml = fullxml_n ""

fullxml_n :: String -> RelExpr
fullxml_n s = filterE (columnT (s ++ "nav.pre") `eqB` columnT (s ++ "txt.pre") `andB` columnT (s ++ "nav.tag") `eqB` columnT (s ++ "tag.tag")) (rawTableR "nav" (s ++ "nav") `cartesianE` rawTableR "txt" (s ++ "txt") `cartesianE` rawTableR "tag" (s ++ "tag"))

transformNodeTest :: XQ.NodeTest -> ICO RelBExpr
transformNodeTest (XQ.NameTest n) = 
    columnT "tag.ttext" `eqB` (stringT $ BSLC8.pack n) `andB` columnT "nav.type" `eqB` intT 1
transformNodeTest (XQ.Wildcard) = trueB
transformNodeTest (XQ.KindTest XQ.AnyKind) = trueB
transformNodeTest (XQ.KindTest XQ.TextKind) = columnT "nav.type" `eqB` intT 2
transformNodeTest (XQ.KindTest XQ.ElementKind) = columnT "nav.type" `eqB` intT 1

transformAxis :: XQ.Axis -> ICO RelBExpr
transformAxis XQ.ChildAxis = columnT "onav.pre" `eqB` columnT "nav.parent"
transformAxis XQ.DescendantAxis = columnT "onav.pre" `addT` intT 1 `leB` columnT "nav.pre" `andB` columnT "onav.pre" `addT` columnT "onav.size" `subT` intT 1 `geB` columnT "nav.pre"

varslist :: UsedColumns a => a -> [String]
varslist = Set.toList . usedVars

n_table_field ('.':r) = ("",r)
n_table_field (c:r) = first (c:) $ n_table_field r

n_table = fst . n_table_field
n_field = snd . n_table_field

-- postać tworzonych wyrażeń:
-- iter l re
-- gdzie (length l) pierwszych kolumn re odpowiada l
-- kolejne kolumny odpowiadają zmiennym i mają dobre nazwy

cshift k = renumber (+ k * num_columns fullxml)

iterTransform :: XQ.ExprSingle RelQuery -> XQ.ExprSingle RelQuery
-- do tego będą potrzebne sprytności w alg.relacji!
iterTransform (XQ.ElemConsExpr n as e) = XQ.ElemConsExpr n as (map iterTransform e)
iterTransform (XQ.BoolExpr op e1 e2) = case (iterTransform e1, iterTransform e2) of
    (XQ.UserExpr (IterN [] [] b1 k1 e1'), XQ.UserExpr (IterN [] [] b2 k2 e2')) -> case op of
        XQ.AndOp -> XQ.UserExpr $ IterN [] [] (b1 ++ cshift k1 b2) (k1 + k2) e1'
--        XQ.OrOp -> XQ.UserExpr $ Iter l1 (r1 `Union` r2) e1'
    (e1', e2') -> XQ.BoolExpr op e1' e2'
iterTransform (XQ.CompExpr op e1 e2) = case (iterTransform e1, iterTransform e2) of
    (XQ.UserExpr (IterN [v1] [pv1] b1 k1 (XQ.VarRefExpr v1')), XQ.UserExpr (IterN [v2] [pv2] b2 k2 (XQ.VarRefExpr v2'))) | v1 == v1' && v2 == v2' ->
        XQ.UserExpr (IterN [] [] (b1 ++ cshift k1 b2 ++ renumber (+pv1) (runICO fullxml $ columnT "nav.type" `eqB` intT 2) ++ cshift k1 (renumber (+pv2) $ runICO fullxml $ columnT "nav.type" `eqB` intT 2) ++ [[CompB EqC (renumber (+pv1) $ runICO fullxml $ columnT "txt.text") (cshift k1 $ renumber (+pv2) $ runICO fullxml $ columnT "txt.text")]]) (k1 + k2) (XQ.FunctionCallExpr "true" []))
    (XQ.UserExpr (IterN [v] [pv] b k e), XQ.StringLiteral s) ->
        XQ.UserExpr (IterN [] [] (b ++ [[CompB EqC (ColumnT pv) (cshift k $ ColumnT 0)]] ++ cshift k (runICO fullxml $ columnT "nav.type" `eqB` intT 2 `andB` columnT "txt.text" `eqB` (stringT $ BSLC8.pack s))) (k + 1) (XQ.FunctionCallExpr "true" []))
    (e1', e2') -> XQ.CompExpr op e1' e2'
-- hacky...
iterTransform (XQ.FunctionCallExpr "true" []) = XQ.UserExpr $ IterN [] [] [[CompB EqC (ColumnT 0) (ResultT $ DbInt 1)]] 1 (XQ.FunctionCallExpr "true" [])
iterTransform (XQ.FunctionCallExpr "false" []) = XQ.UserExpr $ IterN [] [] [[]] 0 (XQ.FunctionCallExpr "false" [])
iterTransform (XQ.FunctionCallExpr n l) = XQ.FunctionCallExpr n (map iterTransform l)
iterTransform (XQ.SequenceExpr e) = XQ.SequenceExpr (map iterTransform e)
iterTransform (XQ.UserExpr (IterN v p b k e)) = XQ.UserExpr (IterN v p b k (iterTransform e))
iterTransform (XQ.IfExpr i t e@(XQ.SequenceExpr [])) = case iterMerge $ iterTransform i of
    XQ.UserExpr (IterN v _ b k i') | Set.fromList v `Set.intersection` XQ.xqfreevars i' == Set.empty -> XQ.UserExpr $ IterN [] [] b k (iterTransform t)
    i' -> XQ.IfExpr i' (iterTransform t) e
iterTransform (XQ.IfExpr i t e) = XQ.IfExpr (iterTransform i) (iterTransform t) (iterTransform e)
iterTransform (XQ.VarRefExpr n) = XQ.UserExpr $ IterN ["x x"] [0] [[CompB EqC (ColumnT 0) (VarT n)]] 1 (XQ.VarRefExpr "x x")
iterTransform (XQ.FLWORExpr [XQ.ForClause [(XQ.VarBinding vn x,x',e)]] Nothing [] f) = case iterTransform e of
    XQ.UserExpr (IterN [v] [pv] b k (XQ.VarRefExpr v')) | v == v' -> 
        XQ.UserExpr $ IterN [vn] [pv] b k (iterTransform f)
    e' -> XQ.FLWORExpr [XQ.ForClause [(XQ.VarBinding vn x,x',e')]] Nothing [] (iterTransform f)
iterTransform (XQ.QuantifiedExpr XQ.SomeQ [(XQ.VarBinding vn x,e)] f) = case iterTransform e of
    XQ.UserExpr (IterN [v] [pv] b k (XQ.VarRefExpr v')) | v == v' -> 
        XQ.UserExpr $ IterN [vn] [pv] b k (iterTransform f)
    e' -> XQ.QuantifiedExpr XQ.SomeQ [(XQ.VarBinding vn x,e')] (iterTransform f)
iterTransform (XQ.PathExpr False [XQ.FilterStep (XQ.VarRefExpr v) [], XQ.AxisStep a t []]) = 
    XQ.UserExpr $ IterN ["x x"] [num_columns fullxml] (runICO ctx $ transformAxis a `andB` transformNodeTest t `andB` columnT "onav.pre" `eqB` varT v) 2 (XQ.VarRefExpr "x x")
    where ctx = fullxml_n "o" `cartesianE` fullxml
iterTransform t = t

iterMerge (XQ.UserExpr (IterN v1 p1 b1 k1 (XQ.UserExpr (IterN v2 p2 b2 k2 e)))) = iterMerge $ 
    XQ.UserExpr (IterN (v1 ++ v2) (p1 ++ cshift k1 p2) b' (k1 + k2) e)
    where 
    b' = b1 ++ subst f (cshift k1 b2)
    f v | Just k <- elemIndex v v1 = ColumnT (p1 !! k)
    f v = VarT v
iterMerge t = inductively iterMerge t

iterOptimize (XQ.UserExpr (Iter v b e)) = 
    XQ.UserExpr (Iter v (optimize b) (iterOptimize e))
iterOptimize t = inductively iterOptimize t

iterGlue (XQ.UserExpr (IterN vs p b k e)) 
    | (u',v',conds):_ <- g =
        let (u, v) = if u' < v' then (u', v') else (v', u')
            p' = renumber g p
            b' = renumber g (filter (`notElem` conds) b)
            g k | k >= v && k < v+xc = u-v+k
                | k >= v+xc = k-xc
                | otherwise = k
        in iterGlue (XQ.UserExpr (IterN vs p' b' (k-1) e)) where
            g = (do
                cond@[CompB EqC (ColumnT u') (ColumnT v')] <- filter f b
                return (u',v',[cond])) `mplus` (do
                cond1@[CompB EqC (ColumnT u') (VarT zz1)] <- filter h b
                cond2@[CompB EqC (ColumnT v') (VarT zz2)] <- filter h b
                guard (zz1 == zz2 && u' < v')
                return (u',v',[cond2]))
            f [CompB EqC (ColumnT u) (ColumnT v)] 
                | u `mod` xc == 0 && v `mod` xc == 0 = True
            f _ = False
            h [CompB EqC (ColumnT u) (VarT _)] 
                | u `mod` xc == 0 = True
            h _ = False
            xc = num_columns fullxml
iterGlue t = inductively iterGlue t

iterConvert (XQ.UserExpr (IterN v p b k e)) = 
    XQ.UserExpr (Iter v (projectR p $ filterR (SimpleFilter b) $ cart !! k) (iterConvert e))
    where
    cart = scanl (\x k -> x `cartesianE` fullxml_n (show k)) (constTableR [] [] [[]] "lol") [0..]
iterConvert t = inductively iterConvert t

inductively f (XQ.UserExpr (IterN v p b k e)) = 
    XQ.UserExpr (IterN v p b k (f e))
inductively f (XQ.UserExpr (Iter v b e)) = 
    XQ.UserExpr (Iter v b (f e))
inductively f (XQ.IfExpr i t e) = 
    XQ.IfExpr (f i) (f t) (f e)
inductively f (XQ.ElemConsExpr n as e) = 
    XQ.ElemConsExpr n as (map f e)
inductively f (XQ.BoolExpr b e1 e2) = 
    XQ.BoolExpr b (f e1) (f e2)
inductively f (XQ.CompExpr b e1 e2) = 
    XQ.CompExpr b (f e1) (f e2)
inductively f (XQ.FunctionCallExpr n es) = 
    XQ.FunctionCallExpr n (map f es) 
inductively f (XQ.SequenceExpr es) = 
    XQ.SequenceExpr (map f es) 
inductively f t = t

varsubst a z (XQ.UserExpr (IterN v p b k e)) 
    | a `notElem` v = XQ.UserExpr (IterN v p (subst (\x -> VarT $ if x == a then z else x) b) k (varsubst a z e))
    | otherwise = XQ.UserExpr (IterN v p b k e)
varsubst a z (XQ.UserExpr (Iter v b e)) 
    | a `notElem` v = XQ.UserExpr (Iter v (subst (\x -> VarT $ if x == a then z else x) b) (varsubst a z e))
    | otherwise = XQ.UserExpr (Iter v b e)
varsubst a z (XQ.VarRefExpr v) 
    | a == v = XQ.VarRefExpr z
    | otherwise = XQ.VarRefExpr v
varsubst a b e = inductively (varsubst a b) e

uniqueVars e = evalState (uv e) 0 where
    uv (XQ.UserExpr (IterN [] p b k e)) = do
        liftM (XQ.UserExpr . IterN [] p b k) $ uv e
    uv (XQ.UserExpr (IterN [v] p b k e)) = do
        z <- get
        put (z+1)
        let v' = v ++ show z
        liftM (XQ.UserExpr . IterN [v'] p (subst (\x -> VarT $ if x == v then v' else x) b) k) $ uv (varsubst v v' e)
    uv (XQ.IfExpr i t e) =
        liftM3 XQ.IfExpr (uv i) (uv t) (uv e)
    uv (XQ.ElemConsExpr n as es) = 
        liftM (XQ.ElemConsExpr n as) $ mapM uv es
    uv (XQ.BoolExpr b e1 e2) = 
        liftM2 (XQ.BoolExpr b) (uv e1) (uv e2)
    uv (XQ.CompExpr b e1 e2) = 
        liftM2 (XQ.CompExpr b) (uv e1) (uv e2)
    uv (XQ.FunctionCallExpr n es) = 
        liftM (XQ.FunctionCallExpr n) $ mapM uv es
    uv (XQ.SequenceExpr es) = 
        liftM XQ.SequenceExpr $ mapM uv es
    uv t = return t

transform = iterOptimize . iterConvert . iterGlue . iterMerge . uniqueVars . iterTransform

