module XQueryVisualize where

import Control.Monad
import Control.Monad.State

import qualified System.IO.UTF8 as SIU

import RelationAlgebraVisualize
import RelationAlgebraStatistics
import qualified XQuery as XQ
import XQuerySemantics(RelQuery(..))

import Debug.Trace

xquery_relexprs xq = execState (xquery_rel_h xq) []

xquery_rel_h (XQ.UserExpr (Iter _ r e)) = do
    modify (r:)
    xquery_rel_h e
xquery_rel_h (XQ.ElemConsExpr _ _ e) = mapM_ xquery_rel_h e
xquery_rel_h (XQ.BoolExpr _ e1 e2) = xquery_rel_h e1 >> xquery_rel_h e2
xquery_rel_h (XQ.CompExpr _ e1 e2) = xquery_rel_h e1 >> xquery_rel_h e2
xquery_rel_h (XQ.IfExpr e1 e2 e3) = xquery_rel_h e1 >> xquery_rel_h e2 >> xquery_rel_h e3
xquery_rel_h (XQ.FunctionCallExpr _ es) = mapM_ xquery_rel_h es
xquery_rel_h (XQ.SequenceExpr es) = mapM_ xquery_rel_h es
xquery_rel_h (XQ.PathExpr _ ses) = mapM_ xquery_path_rel_h ses
xquery_rel_h (XQ.QuantifiedExpr _ [(XQ.VarBinding _ _, e)] c) =
    xquery_rel_h e >> xquery_rel_h c
xquery_rel_h (XQ.FLWORExpr [XQ.ForClause [(XQ.VarBinding _ _,_,e)]] _ _ r) =
    xquery_rel_h e >> xquery_rel_h r
xquery_rel_h e = return ()

xquery_path_rel_h (XQ.AxisStep _ _ _) = return ()
xquery_path_rel_h (XQ.FilterStep f _) = xquery_rel_h f

debug_xquery q s = mapM_ h (zip [1..] (xquery_relexprs q)) where
    h (k, r) = do
        SIU.writeFile ("rq" ++ show k ++ ".dot") (relexpr_to_dot r)
        print k
        putStr ("rowcount: " ++ show (runWithStats (rowcount r) s) ++ "\n")
        putStr ("opcount: " ++ show (runWithStats (opcount_execute r) s) ++ "\n")


