module XmlBS(Document(..), Element(..), Attribute, Content(..), Prolog(..), Misc(..), Reference(..), AttValue(..), CP(..), ContentSpec(..), Modifier(..), XMLDecl(..), AttType(..), AttListDecl(..), AttDef(..), EntityDecl(..), info, --setInfo,
    documentToXmlBS, contentToXmlBS,
    documentFromXmlBS, contentFromXmlBS) where

import Text.XML.HaXml(Prolog(..), SymTab, EntityDef, Misc(..), Reference(..), CP(..), ContentSpec(..), Modifier(..), XMLDecl(..), AttType(..), AttListDecl(..), AttDef(..), EntityDecl(..))
import qualified Text.XML.HaXml.Types as XML
import qualified Data.ByteString.Lazy as BSL
import qualified Data.ByteString.Lazy.Char8 as BSLC8

data Document i = Document Prolog (SymTab EntityDef) (Content i) [Misc]
    deriving Eq

data Content i = Elem (Element i) i deriving Eq
type Attribute = (Name, AttValue)
data Element i = CElem Name [Attribute] [Content i]
               | CString Bool CharData
                        -- ^ bool is whether whitespace is significant
               | CRef Reference
               | CMisc Misc
    deriving Eq

info :: Content t -> t
info (Elem (CElem _ _ _) i) = i
info (Elem (CString _ _) i) = i
info (Elem (CRef _) i) = i
info (Elem (CMisc _) i) = i
{-
setInfo i (CElem e _) = CElem e i
setInfo i (CString b e _) = CString b e i
setInfo i (CRef e _) = CRef e i
setInfo i (CMisc e _) = CMisc e i
-}
type Name     = BSL.ByteString       -- non-empty string
type Names    = [Name]           -- non-empty list
data AttValue    = AttValue [Either BSL.ByteString Reference] deriving Eq

type CharData = BSL.ByteString

{-
data Reference    = RefEntity EntityRef
                  | RefChar CharRef
                  deriving (Eq,Show)
type EntityRef    = Name
type CharRef      = Int
-}

documentToXmlBS :: XML.Document a -> Document a
documentToXmlBS (XML.Document p e c m) = Document p e (elementToXmlBS c) m

elementToXmlBS :: XML.Element a -> Content a
elementToXmlBS e = contentToXmlBS (XML.CElem e undefined)

contentToXmlBS :: XML.Content a -> Content a
contentToXmlBS (XML.CElem (XML.Elem n a cs) i) = Elem (CElem (BSLC8.pack n) [] (map contentToXmlBS cs)) i
contentToXmlBS (XML.CString b c i) = Elem (CString b (BSLC8.pack c)) i
contentToXmlBS (XML.CRef r i) = Elem (CRef r) i
contentToXmlBS (XML.CMisc m i) = Elem (CMisc m) i

documentFromXmlBS :: Document a -> XML.Document a
documentFromXmlBS (Document p e c m) = XML.Document p e (elementFromXmlBS c) m

elementFromXmlBS :: Content a -> XML.Element a
elementFromXmlBS e = i where XML.CElem i _ = contentFromXmlBS e

contentFromXmlBS :: Content a -> XML.Content a
contentFromXmlBS (Elem (CElem n a cs) i) = XML.CElem (XML.Elem (BSLC8.unpack n) [] (map contentFromXmlBS cs)) i
contentFromXmlBS (Elem (CString b c) i) = XML.CString b (BSLC8.unpack c) i
contentFromXmlBS (Elem (CRef r) i) = XML.CRef r i
contentFromXmlBS (Elem (CMisc m) i) = XML.CMisc m i


